package j2d.gui.menu;

import gui.run.RunMenu;
import gui.run.RunMenuItem;

import java.awt.*;

import j2d.ImageUtils;
import j2d.ShortImageBean;
import ip.transforms.Kernels;
import math.fourierTransforms.r2.FFT2dComplex;
import math.fourierTransforms.pfa.PFAImage;

/**
 * Copyright 2005, DocJava, Inc.
 * User: dlyon
 * Date: Nov 2, 2005
 * Time: 5:47:28 PM
 *
 * Anne MacIlvain - Modified to do correlation, for the final.
 * Notes:
 * 1) I've tried several variations on a theme, and I'm leaving all of the
 * code intact, even if I end up not using it, so you'll see
 * that I have expended time and energy on this.
 * 2) My final incorporates 4 methods of convolution:
 *    1) FFT
 *    2) PFA
 *    3) Time domain - based on formula from internet link
 *    4) Time domain Try2 (which is really a variation on the
 *    FFT method, but which extracts just the template, then
 *    zero pads it out to match the image)
 * 3) To verify the algorithms, I first used the "UTK" image (which
 * I created and have included as part of my final) then used the baboon.
 * I tried the space shuttle, but it didn't correlate well, no matter
 * which method was applied to it.
 * 4) I believe the FFT and PFA methods are working, but the Time
 * domain and Time domain Try2 methods don't work. Extensive debugging
 * has failed to find the bugs and fix them. :-(
 *
*/

public class FilterMenu {
    public static RunMenu getFilterMenu(final j2d.gui.Main main) {
        RunMenu rm = new RunMenu("F[ilter");
        rm.add(EdgeMenu.getEdgeMenu(main));
        rm.add(BlurMenu.getBlurMenu(main));

        rm.add(getHpMenu(main));
        rm.add(getLpMenu(main));
        rm.add(getMedianMenu(main));
        rm.add(getOtherTransformsMenu(main));
        rm.add(
                new RunMenuItem("BorderPanel") {
                    public void run() {
                        main.addPanel(
                                new j2d.border.BorderPanel(main),
                                getText());
                    }
                });
        rm.add(
                new RunMenuItem("Convolution") {
                    public void run() {
                        main.launchConvolutionControls();
                        //     Image img = main.getImage();
                        //     bufferedImageConvolution.ImageMatrix im=new bufferedImageConvolution.ImageMatrix(img);
                        //     main.setImage(im.getImage());
                    }
                });
        rm.add(new RunMenuItem("PFA Correlation"){
            public void run(){
                //todo this is your final....right here...
                Image img1 = main.getImage();
                Image img2 = main.getSecondImage();
                //todo this is not done, it is just a quick prototype
                //ImageUtils.displayImage(img1,"img1");
                //ImageUtils.displayImage(ImageUtils.upsideDownMirror(img2),"img2");

                // Anne MacIlvain
                Image nccFFTImage = normalizedCrossCorrelationFFT(img1,img2);
                ImageUtils.displayImage(nccFFTImage,"scaled result");

                // Anne MacIlvain
                int tryPFA = 1;
                if (tryPFA > 0) {
                    Image nccPFAImage = normalizedCrossCorrelationPFA(img1,img2);
                    ImageUtils.displayImage(ImageUtils.quadSwap(nccPFAImage),"PFA result");
                }

                // Anne MacIlvain
                // Try formula from website link.
                int tryWebsiteLink = 1;
                if (tryWebsiteLink > 0) {
                    Image nccImage = normalizedCrossCorrelation(img1,img2);
                    ImageUtils.displayImage(nccImage,"time domain result");
                    Image nccTry2Image = NCCTry2(img1,img2);
                    ImageUtils.displayImage(nccTry2Image,"time domain result (try2)");
                }

                main.addImage(ImageUtils.quadSwap(nccFFTImage));
            }
        });
          rm.add(new RunMenuItem("spatial Correlation"){
            public void run(){

               //todo this is your final....right here...
                Image img1 = main.getImage();
                Image img2 = main.getSecondImage();
                long startTime = System.currentTimeMillis();
                Image image = ImageUtils.getSpatialCorrelation(img1, img2);
                long elapsedTime = System.currentTimeMillis() - startTime;
                System.out.println("getSpatialCorrelation time in ms:"+elapsedTime);
                main.addImage(image);
            }
        });
        return rm;
    }

    // ===========================================
    // Start of my code additions:
    // ===========================================

    // Anne MacIlvain added:
    private static Image normalizedCrossCorrelationFFT(Image img1, Image img2) {
        // Anne MacIlvain
        // The correlation of a[n] with b[n] is
        // a[n] * b[-n] = c[n], in the time domain.
        // "That is, flipping b[n] left-for-right is
        // accomplished by reversing the sign of the
        // index, i.e., b [-n].
        // So flip the image in the time domain, before
        // working in the frequency domain.
        Image img2Flipped = flip(img2);
        ImageUtils.displayImage(img2Flipped,"flipped image 2");

        // Subtract the mean from each image.
        Image img1MeanRemoved = subtractMean(img1);
        Image img2FlippedMeanRemoved = subtractMean(img2Flipped);

        ImageUtils.displayImage(img1MeanRemoved,"image 1 mean removed");
        ImageUtils.displayImage(img2FlippedMeanRemoved,"image 2 flipped, mean removed");

        // Anne MacIlvain
        // In the frequency domain, a[n] * b[-n] = c[n]
        // is equivalent to a[n] x b[-n] = c[n]
        FFT2dComplex f1 = new FFT2dComplex(img1);
        FFT2dComplex f2 = new FFT2dComplex(img2Flipped);

        f1.fft();
        f2.fft();
        // The conjugate does the same thing as the flip
        // done above, so don't call it.
        //f2.conjugate();

        // Anne MacIlvain
        // Remove mean. In the frequency domain, this is
        // done by zeroing the DC component, which is
        // in the 0 location.
        int removeMean = 0;
        if (removeMean > 0) {
            f1.removeMean();
            f2.removeMean();
        }
        int removeDCBias = 1;
        if (removeDCBias > 0) {
            f1.removeDCBias();
            f2.removeDCBias();
        }

        // todo to PCC or NCC that is the question?
        // Would it have been better to edge detect first,
        // as LOG said?
        f1.mult(f2);
        f1.ifft();

        return f1.getScaledImage();
    }

    // Anne MacIlvain added:
    private static Image normalizedCrossCorrelationPFA(Image img1, Image img2) {
        // Anne MacIlvain
        // The correlation of a[n] with b[n] is
        // a[n] * b[-n] = c[n], in the time domain.
        // "That is, flipping b[n] left-for-right is
        // accomplished by reversing the sign of the
        // index, i.e., b [-n].
        // So flip the image in the time domain, before
        // working in the frequency domain.
        Image img2Flipped = flip(img2);

        PFAImage p1 = new PFAImage(img1);
        PFAImage p2 = new PFAImage(img2Flipped);

        p1.fft();
        p2.fft();

        p1.multiply(p2);

        p1.ifft();

        return p1.getImage();
    }

    // Anne MacIlvain added:
    private static Image normalizedCrossCorrelation(Image img1, Image img2) {
        int display = 0;
        int display2 = 0;
        int img1SubtractMean = 1;
        int templateSubtractMean = 1;
        int img1SubtractMeanEntireImage = 0;
        int flipTemplate = 0;

        ShortImageBean sibCorr = new ShortImageBean(img1);

        // Extract only the template out of the image.
        Image template = extractTemplate(img2);
        ImageUtils.displayImage(template,"template only");

        if (templateSubtractMean > 0) {
            template = subtractMean(template);
            ImageUtils.displayImage(template,"template only mean removed");
        }

        if (flipTemplate > 0) {
            template = flip(template);
            ImageUtils.displayImage(template,"template only flipped");
        }

        if (img1SubtractMeanEntireImage > 0) {
            img1 = subtractMean(img1);
            ImageUtils.displayImage(img1,"image1 mean removed");
        }

        // Get the RGB values for img1
        ShortImageBean sib1 = new ShortImageBean(img1);
        int w1 = sib1.getWidth();
        int h1 = sib1.getHeight();
        float[][] r1 = new float[w1][h1];
        float[][] g1 = new float[w1][h1];
        float[][] b1 = new float[w1][h1];
        r1 = getSIBR(sib1);
        g1 = getSIBG(sib1);
        b1 = getSIBB(sib1);

        short[][] rc = new short[w1][h1];
        short[][] gc = new short[w1][h1];
        short[][] bc = new short[w1][h1];

        // Get the RGB values for img2
        ShortImageBean sib2 = new ShortImageBean();
        sib2 = new ShortImageBean(template);

        int w2 = sib2.getWidth();
        int h2 = sib2.getHeight();
        float[][] r2 = new float[w2][h2];
        float[][] g2 = new float[w2][h2];
        float[][] b2 = new float[w2][h2];
        r2 = getSIBR(sib2);
        g2 = getSIBG(sib2);
        b2 = getSIBB(sib2);

        float[][] r1mean = new float[w2][h2];
        float[][] g1mean = new float[w2][h2];
        float[][] b1mean = new float[w2][h2];

        float[][] rcorr = new float[w1][h1];
        float[][] gcorr = new float[w1][h1];
        float[][] bcorr = new float[w1][h1];

        float rnumer;
        float gnumer;
        float bnumer;
        float rdenom1;
        float gdenom1;
        float bdenom1;
        float rdenom2;
        float gdenom2;
        float bdenom2;

        float zeroGuard = Float.MAX_VALUE;

        // Perform summation and normalization.
        display = 0;
        display2 = 0;
        for (int uu = 0; uu < (w1-w2); uu++) {
            for (int vv = 0; vv < (h1-h2); vv++) {
                // Get the local image mean at location (u,v) -
                // the mean of the image underneath the template
                // whose top left corner lies on pixel (u,v)
                if ( (img1SubtractMean > 0) &&
                        (img1SubtractMeanEntireImage <= 0) ) {
                    r1mean = subtractMean(r1,uu,vv,w2,h2);
                    g1mean = subtractMean(g1,uu,vv,w2,h2);
                    b1mean = subtractMean(b1,uu,vv,w2,h2);
                }
                else {
                    r1mean = getLocalImage(r1,uu,vv,w2,h2);
                    g1mean = getLocalImage(g1,uu,vv,w2,h2);
                    b1mean = getLocalImage(b1,uu,vv,w2,h2);
                }
                rnumer = 0.0f;
                gnumer = 0.0f;
                bnumer = 0.0f;
                rdenom1 = 0.0f;
                gdenom1 = 0.0f;
                bdenom1 = 0.0f;
                rdenom2 = 0.0f;
                gdenom2 = 0.0f;
                bdenom2 = 0.0f;
                for (int xx = 0; xx < w2; xx++) {
                    for (int yy = 0; yy < h2; yy++) {
                        if (display2 < 10) {
                            //displayImage(r1mean,g1mean,b1mean,"local image "+display2);
                            System.out.println("r1,r2 " + r1mean[xx][yy] + " " + r2[xx][yy]);
                            display2 = display2 + 1;
                        }
                        rnumer += r1mean[xx][yy]*r2[xx][yy];
                        gnumer += g1mean[xx][yy]*g2[xx][yy];
                        bnumer += b1mean[xx][yy]*b2[xx][yy];
                        rdenom1 += r1mean[xx][yy]*r1mean[xx][yy];
                        gdenom1 += g1mean[xx][yy]*g1mean[xx][yy];
                        bdenom1 += b1mean[xx][yy]*b1mean[xx][yy];
                        rdenom2 += r2[xx][yy]*r2[xx][yy];
                        gdenom2 += g2[xx][yy]*g2[xx][yy];
                        bdenom2 += b2[xx][yy]*b2[xx][yy];
                    }
                }
                if (display < 10) {
                    System.out.println("r/d1,d2 " + rnumer +
                            " " + rdenom1 + " " + rdenom2);
                    display = display + 1;
                }

                // Guard against division by 0.
                if ( (rdenom1 == 0.0f) || (rdenom2 == 0.0f) ) {
                    rcorr[uu][vv] = zeroGuard;
                }
                else {
                    rcorr[uu][vv] = rnumer/(rdenom1*rdenom2);
                }

                if ( (gdenom1 == 0.0f) || (gdenom2 == 0.0f) ) {
                    gcorr[uu][vv] = zeroGuard;
                }
                else {
                    gcorr[uu][vv] = gnumer/(gdenom1*gdenom2);
                }

                if ( (bdenom1 == 0.0f) || (bdenom2 == 0.0f) ) {
                    bcorr[uu][vv] = zeroGuard;
                }
                else {
                    bcorr[uu][vv] = bnumer/(bdenom1*bdenom2);
                }
            }
        }

        // Scale the result.
        float max = getMax(rcorr,gcorr,bcorr);
        float sf = 255f / max;
        System.out.println("max/sf = " + max + " " + sf);

        display = 0;
        for (int uu = 0; uu < w1; uu++) {
            for (int vv = 0; vv < h1; vv++) {
                if (display < 10) {
                    System.out.println("RGB = " + rcorr[uu][vv] +
                            " " + gcorr[uu][vv] + " " + bcorr[uu][vv]);
                    display = display + 1;
                }
                rc[uu][vv] = (short) (sf*rcorr[uu][vv]);
                gc[uu][vv] = (short) (sf*gcorr[uu][vv]);
                bc[uu][vv] = (short) (sf*bcorr[uu][vv]);
            }
        }

        sibCorr.setR(rc);
        sibCorr.setG(gc);
        sibCorr.setB(bc);

        return sibCorr.getImage();
    }

    // Anne MacIlvain added:
    private static Image NCCTry2(Image img1, Image img2) {

        // Extract only the template out of the image.
        Image template = extractTemplate(img2);
        ImageUtils.displayImage(template,"template only");

        template = flip(template);
        ImageUtils.displayImage(template,"template only flipped");

        template = subtractMean(template);
        ImageUtils.displayImage(template,"template only flipped, mean subtracted");

        // Zero pad the template out to match image of interest,
        // to be able to do the FFT on it.
        Image templatePadded = zeroPadImage(template,img1);
        ImageUtils.displayImage(templatePadded,"template zero padded");

        // Subtract the mean from image of interest.
        Image img1MeanRemoved = subtractMean(img1);

        // In the frequency domain, a[n] * b[-n] = c[n]
        // is equivalent to a[n] x b[-n] = c[n]
        FFT2dComplex f1 = new FFT2dComplex(img1MeanRemoved);
        FFT2dComplex f2 = new FFT2dComplex(templatePadded);

        f1.fft();
        f2.fft();

        // Remove mean. In the frequency domain, this is
        // done by zeroing the DC component, which is
        // in the 0 location.
        int removeMean = 0;
        if (removeMean > 0) {
            f1.removeMean();
            f2.removeMean();
        }
        int removeDCBias = 1;
        if (removeDCBias > 0) {
            f1.removeDCBias();
            f2.removeDCBias();
        }

        // todo to PCC or NCC that is the question?
        // Would it have been better to edge detect first,
        // as LOG said?
        f1.mult(f2);
        f1.ifft();

        return f1.getScaledImage();
    }

    // Anne MacIlvain added:
    private static Image zeroPadImage(Image template, Image img1) {
        ShortImageBean sibImage = new ShortImageBean(img1);
        ShortImageBean sibTemplate = new ShortImageBean(template);
        int w = sibImage.getWidth();
        int h = sibImage.getHeight();
        int wt = sibTemplate.getWidth();
        int ht = sibTemplate.getHeight();
        short[][] rout = new short[w][h];
        short[][] gout = new short[w][h];
        short[][] bout = new short[w][h];
        short[][] rt = new short[wt][ht];
        short[][] gt = new short[wt][ht];
        short[][] bt = new short[wt][ht];
        rt = sibTemplate.getR();
        gt = sibTemplate.getG();
        bt = sibTemplate.getB();
        // Zero fill entire image first.
        for (int xx = 0; xx < w; xx++) {
            for (int yy = 0; yy < h; yy++) {
                rout[xx][yy] = 0;
                gout[xx][yy] = 0;
                bout[xx][yy] = 0;
            }
        }

        int splitImage = 0;

        if (splitImage > 0) {
            // Split template across four corners.
            for (int xx = 0; xx < wt; xx++) {
                for (int yy = 0; yy < ht; yy++) {
                    if ( (xx <= (wt/2-1)) && (yy <= (ht/2-1)) ) {
//                    System.out.println("" + xx + " " + yy + " " +
//                            xx+wt/2 + " " + yy+ht/2 + " " + wt + " " + ht );
                        rout[xx][yy] = rt[xx+wt/2][yy+ht/2];
                        gout[xx][yy] = gt[xx+wt/2][yy+ht/2];
                        bout[xx][yy] = bt[xx+wt/2][yy+ht/2];
                    }
                    else if ( (xx >= wt/2) && (yy <= (ht/2-1)) ) {
                        rout[w-xx+wt/2-1][yy] = rt[xx-wt/2][yy+ht/2];
                        gout[w-xx+wt/2-1][yy] = gt[xx-wt/2][yy+ht/2];
                        bout[w-xx+wt/2-1][yy] = bt[xx-wt/2][yy+ht/2];
                    }
                    else if ( (xx <= (wt/2-1)) && (yy >= ht/2) ) {
                        //System.out.println("" + xx + " " + (h-yy+ht/2) + " " +
                        //        " " + w + " " + h );
                        rout[xx][h-yy+ht/2-1] = rt[xx+wt/2][yy];
                        gout[xx][h-yy+ht/2-1] = gt[xx+wt/2][yy];
                        bout[xx][h-yy+ht/2-1] = bt[xx+wt/2][yy];
                    }
                    else if ( (xx >= wt/2) && (yy >= ht/2) ) {
                        rout[w-xx+wt/2-1][h-yy+ht/2-1] = rt[xx-wt/2][yy-ht/2];
                        gout[w-xx+wt/2-1][h-yy+ht/2-1] = gt[xx-wt/2][yy-ht/2];
                        bout[w-xx+wt/2-1][h-yy+ht/2-1] = bt[xx-wt/2][yy-ht/2];
                    }
                }
            }
        }
        else {
            for (int xx = 0; xx < wt; xx++) {
                for (int yy = 0; yy < ht; yy++) {
                    rout[xx][yy] = rt[xx][yy];
                    gout[xx][yy] = gt[xx][yy];
                    bout[xx][yy] = bt[xx][yy];
                }
            }
        }

        sibImage.setR(rout);
        sibImage.setG(gout);
        sibImage.setB(bout);
        return sibImage.getImage();
    }

    // Anne MacIlvain added:
    private static void displayImage(float[][] r, float[][] g, float[][] b, String title) {
        //To change body of created methods use File | Settings | File Templates.
        ShortImageBean sib = new ShortImageBean();
        int w = r.length;
        int h = r[0].length;
        short[][] rs = new short[w][h];
        short[][] gs = new short[w][h];
        short[][] bs = new short[w][h];
        for (int xx = 0;xx < w; xx++) {
            for (int yy = 0; yy < h; yy++) {
                rs[xx][yy] = (short) r[xx][yy];
                gs[xx][yy] = (short) g[xx][yy];
                bs[xx][yy] = (short) b[xx][yy];
            }
        }
        sib.setR(rs);
        sib.setG(gs);
        sib.setB(bs);

        ImageUtils.displayImage(sib.getImage(), title);
    }

    // Anne MacIlvain added:
    private static float[][] getSIBR(ShortImageBean sib) {
        int w = sib.getWidth();
        int h = sib.getHeight();
        short[][] rin = new short[w][h];
        float[][] rout = new float[w][h];
        rin = sib.getR();
        for (int xx = 0; xx < w; xx++) {
            for (int yy = 0; yy < h; yy++) {
                rout[xx][yy] = rin[xx][yy];
            }
        }
        return rout;
    }

    // Anne MacIlvain added:
    private static float[][] getSIBG(ShortImageBean sib) {
        int w = sib.getWidth();
        int h = sib.getHeight();
        short[][] rin = new short[w][h];
        float[][] rout = new float[w][h];
        rin = sib.getG();
        for (int xx = 0; xx < w; xx++) {
            for (int yy = 0; yy < h; yy++) {
                rout[xx][yy] = rin[xx][yy];
            }
        }
        return rout;
    }

    // Anne MacIlvain added:
    private static float[][] getSIBB(ShortImageBean sib) {
        int w = sib.getWidth();
        int h = sib.getHeight();
        short[][] rin = new short[w][h];
        float[][] rout = new float[w][h];
        rin = sib.getB();
        for (int xx = 0; xx < w; xx++) {
            for (int yy = 0; yy < h; yy++) {
                rout[xx][yy] = rin[xx][yy];
            }
        }
        return rout;
    }

    // Anne MacIlvain added:
    private static float[][] subtractMean(float[][] r, int uu, int vv, int w, int h) {
        float[][] localMean = new float[w][h];

        float rmean = 0.0f;

        rmean = 0.0f;

        // Get the average (mean) value.
        for (int ww = 0; ww < w; ww++) {
            for (int hh = 0; hh < h; hh++) {
                rmean += r[uu+ww][vv+hh];
            }
        }

        rmean = rmean / (w*h);

        // Subtract out the average (mean) value.
        for (int ww = 0; ww < w; ww++) {
            for (int hh = 0; hh < h; hh++) {
                localMean[ww][hh] = r[uu+ww][vv+hh] - rmean;
            }
        }

        return localMean;
    }

    // Anne MacIlvain added:
    private static float[][] getLocalImage(float[][] r, int uu, int vv, int w, int h) {
        float[][] localImage = new float[w][h];

        // Copy sub image.
        for (int ww = 0; ww < w; ww++) {
            for (int hh = 0; hh < h; hh++) {
                localImage[ww][hh] = r[uu+ww][vv+hh];
            }
        }

        return localImage;
    }

    // Anne MacIlvain added:
    private static Image extractTemplate(Image img) {
        // ========================================================
        // Algorithm:
        // ========================================================
        // Scan each row until a non black pixel is found.
        // Keep track of the leftTop,rightTop,leftBottom,rightBottom
        // pixel positions found that contain non black.
        // When another full row of black is found, stop.
        // This will create a rectangle template, but within it
        // should be the shape the user applied to the image.
        // ========================================================
        // Get the RGB values for img
        ShortImageBean sib = new ShortImageBean(img);
        int w = sib.getWidth();
        int h = sib.getHeight();
        short[][] r = new short[w][h];
        short[][] g = new short[w][h];
        short[][] b = new short[w][h];
        r = sib.getR();
        g = sib.getG();
        b = sib.getB();
        int leftTop = h;
        int leftBottom = 0;
        int rightTop = w;
        int rightBottom = 0;
        int rowBlackCount;
        boolean nonBlackFound = false;
        for (int x = 0; x < w; x++) {
            rowBlackCount = 0;
            for (int y = 0; y < h; y++) {
                if ( (r[x][y] == 0) && (g[x][y] == 0) && (b[x][y] == 0)) {
                    rowBlackCount = rowBlackCount + 1;
                }
                else {
                    nonBlackFound = true;
                    //System.out.println("x,y = " + x + " " + y);
                    if (x < rightTop) rightTop = x;
                    if (x > rightBottom) rightBottom = x;
                    if (y < leftTop) leftTop = y;
                    if (y > leftBottom) leftBottom = y;
                }
            } // end for (int y = 0; y < h; y++) {

            if ((nonBlackFound) && (rowBlackCount == w)) break;

        } // end for (int x = 0; x < w; x++) {

        // Extract template
        int wt = rightBottom - rightTop + 1;
        int ht = leftBottom - leftTop + 1;
        System.out.println("w,h = " + wt + " " + ht);
        System.out.println("right B,T = " + rightBottom + " " + rightTop);
        System.out.println("left  B,T = " + leftBottom + " " + leftTop);
        short[][] rt = new short[wt][ht];
        short[][] gt = new short[wt][ht];
        short[][] bt = new short[wt][ht];
        for (int x = 0; x < wt; x++) {
            for (int y = 0; y < ht; y++) {
                rt[x][y] = r[x+rightTop][y+leftTop];
                gt[x][y] = g[x+rightTop][y+leftTop];
                bt[x][y] = b[x+rightTop][y+leftTop];
            }
        }

        ShortImageBean sibTemplate = new ShortImageBean();
        sibTemplate.setR(rt);
        sibTemplate.setG(gt);
        sibTemplate.setB(bt);

        return sibTemplate.getImage();
    }

    // Anne MacIlvain added:
    private static Image flip(Image img) {
        ShortImageBean sib = new ShortImageBean(img);
        int w = sib.getWidth();
        int h = sib.getHeight();
        int wflip = w - 1;
        int hflip = h - 1;
        short[][] r = new short[w][h];
        short[][] g = new short[w][h];
        short[][] b = new short[w][h];
        short[][] rflip = new short[w][h];
        short[][] gflip = new short[w][h];
        short[][] bflip = new short[w][h];

        // Get the RGB values.
        r = sib.getR();
        g = sib.getG();
        b = sib.getB();

        // Flip the RGB values.
        for (int ww = 0; ww < w; ww++) {
            for (int hh = 0; hh < h; hh++) {
                rflip[wflip-ww][hflip-hh] = r[ww][hh];
                gflip[wflip-ww][hflip-hh] = g[ww][hh];
                bflip[wflip-ww][hflip-hh] = b[ww][hh];
            }
        }

        // Set the RGB values back into the ShortImageBean.
        sib.setR(rflip);
        sib.setG(gflip);
        sib.setB(bflip);

        // Return the flipped image.
        return sib.getImage();
    }

    // Anne MacIlvain added:
    private static Image subtractMean(Image img) {
        ShortImageBean sib = new ShortImageBean(img);
        int w = sib.getWidth();
        int h = sib.getHeight();
        short[][] r = new short[w][h];
        short[][] g = new short[w][h];
        short[][] b = new short[w][h];
        float rmean = 0.0f;
        float gmean = 0.0f;
        float bmean = 0.0f;

        // Get the RGB values.
        r = sib.getR();
        g = sib.getG();
        b = sib.getB();

        rmean = 0.0f;
        gmean = 0.0f;
        bmean = 0.0f;

        // Get the average (mean) values.
        for (int ww = 0; ww < w; ww++) {
            for (int hh = 0; hh < h; hh++) {
                rmean += r[ww][hh];
                gmean += g[ww][hh];
                bmean += b[ww][hh];
            }
        }

        rmean = rmean / (w*h);
        gmean = gmean / (w*h);
        bmean = bmean / (w*h);

        // Subtract out the average (mean) values.
        for (int ww = 0; ww < w; ww++) {
            for (int hh = 0; hh < h; hh++) {
                r[ww][hh] -= rmean;
                g[ww][hh] -= gmean;
                b[ww][hh] -= bmean;
            }
        }

        // Set the RGB values back into the ShortImageBean.
        sib.setR(r);
        sib.setG(g);
        sib.setB(b);

        // Return the flipped image.
        return sib.getImage();
    }

    // Anne MacIlvain added:
    private static float getMax(float[][] rcorr, float[][] gcorr, float[][] bcorr) {
        float rmax = getMax(rcorr);
        float gmax = getMax(gcorr);
        float bmax = getMax(bcorr);
        float max = rmax;
        if (max < gmax) max = gmax;
        if (max < bmax) max = bmax;
        return max;
    }

    // Anne MacIlvain added:
    private static float getMax(float[][] array) {
        float max = Float.MIN_VALUE;
        for (int rr = 0; rr < array.length; rr++) {
            for (int cc = 0; cc < array[0].length; cc++) {
                if (array[rr][cc] > max)
                    max = array[rr][cc];
            }
        }
        return max;
    }
    // ==========================================================
    // end of Anne MacIlvain added.
    // ==========================================================


    public static RunMenu getHpMenu(final j2d.gui.Main main) {
        RunMenu rm = new RunMenu("Hp");
        rm.add(
                new RunMenuItem("Hp1") {
                    public void run() {
                        Image img = main.getImage();
                        img
                                = ImageUtils.convolution(
                                        img, Kernels.getHp1());
                        main.setImage(img);
                    }
                });
        rm.add(
                new RunMenuItem("Hp2") {
                    public void run() {
                        Image img = main.getImage();
                        img
                                = ImageUtils.convolution(
                                        img, Kernels.getHp2());
                        main.setImage(img);
                    }
                });
        rm.add(
                new RunMenuItem("Hp3") {
                    public void run() {
                        Image img = main.getImage();
                        img
                                = ImageUtils.convolution(
                                        img, Kernels.getHp3());
                        main.setImage(img);
                    }
                });
        rm.add(
                new RunMenuItem("Hp4") {
                    public void run() {
                        Image img = main.getImage();
                        img
                                = ImageUtils.convolution(
                                        img, Kernels.getHp4());
                        main.setImage(img);
                    }
                });
        rm.add(
                new RunMenuItem("Hp5") {
                    public void run() {
                        Image img = main.getImage();
                        img
                                = ImageUtils.convolution(
                                        img, Kernels.getHp5());
                        main.setImage(img);
                    }
                });

        return rm;
    }

    public static RunMenu getLpMenu(final j2d.gui.Main main) {
        RunMenu rm = new RunMenu("Lp");
        rm.add(
                new RunMenuItem("Lp1") {
                    public void run() {
                        Image img = main.getImage();
                        img
                                = ImageUtils.convolution(
                                        img, Kernels.getLp1());
                        main.setImage(img);
                    }
                });
        rm.add(
                new RunMenuItem("Lp2") {
                    public void run() {
                        Image img = main.getImage();
                        img
                                = ImageUtils.convolution(
                                        img, Kernels.getLp2());
                        main.setImage(img);
                    }
                });
        rm.add(
                new RunMenuItem("Lp3") {
                    public void run() {
                        Image img = main.getImage();
                        img
                                = ImageUtils.convolution(
                                        img, Kernels.getLp3());
                        main.setImage(img);
                    }
                });

        return rm;
    }

    public static RunMenu getOtherTransformsMenu(final j2d.gui.Main main) {
        RunMenu rm = new RunMenu("Other Filters");
        rm.add(
                new RunMenuItem("Average 3x3") {
                    public void run() {
                        Image img = main.getImage();
                        img
                                = ImageUtils.convolution(
                                        img, Kernels.getAverage3x3());
                        main.setImage(img);
                    }
                });
        rm.add(
                new RunMenuItem("Gabor 7") {
                    public void run() {
                        Image img = main.getImage();
                        img
                                = ImageUtils.convolution(
                                        img, Kernels.getGabor7());
                        main.setImage(img);
                    }
                });
        rm.add(
                new RunMenuItem("Hat 13") {
                    public void run() {
                        Image img = main.getImage();
                        img
                                = ImageUtils.convolution(
                                        img, Kernels.getHat13());
                        main.setImage(img);
                    }
                });
        rm.add(
                new RunMenuItem("Mean 3") {
                    public void run() {
                        Image img = main.getImage();
                        img
                                = ImageUtils.convolution(
                                        img, Kernels.getMean3());
                        main.setImage(img);
                    }
                });
        rm.add(
                new RunMenuItem("Mean 9") {
                    public void run() {
                        Image img = main.getImage();
                        img
                                = ImageUtils.convolution(
                                        img, Kernels.getMean9());
                        main.setImage(img);
                    }
                });
        rm.add(
                new RunMenuItem("Sobel") {
                    public void run() {
                        Image img = main.getImage();
                        img
                                = ImageUtils.convolution(
                                        img, Kernels.getSobel());
                        main.setImage(img);
                    }
                });

        return rm;
    }

    public static float[][] shortMatrixToFloat(short sm[][]) {
        float[][] fm = new float[sm.length][sm[0].length];

        for (int x = 0; x < sm.length; x++)
            System.arraycopy(sm[x], 0, fm[x], 0, sm[x].length);
        return fm;
    }

    public static RunMenu getMedianMenu(final j2d.gui.Main main) {
        RunMenu rm = new RunMenu("Median Filters");
        rm.add(
                new RunMenuItem("Median 1x2") {
                    public void run() {
                        Image img = main.getImage();
                        img
                                = ImageUtils.convolution(
                                        img,
                                        shortMatrixToFloat(
                                                Kernels.getMedian1x2()));
                        main.setImage(img);
                    }
                });
        rm.add(
                new RunMenuItem("Median 2x1") {
                    public void run() {
                        Image img = main.getImage();
                        img
                                = ImageUtils.convolution(
                                        img,
                                        shortMatrixToFloat(
                                                Kernels.getMedian2x1()));
                        main.setImage(img);
                    }
                });
        rm.add(
                new RunMenuItem("Median Cross 3x3") {
                    public void run() {
                        Image img = main.getImage();
                        img
                                = ImageUtils.convolution(
                                        img,
                                        shortMatrixToFloat(
                                                Kernels.getMedianCross3x3()));
                        main.setImage(img);
                    }
                });
        rm.add(
                new RunMenuItem("Median Cross 5x5") {
                    public void run() {
                        Image img = main.getImage();
                        img
                                = ImageUtils.convolution(
                                        img,
                                        shortMatrixToFloat(
                                                Kernels.getMedianCross5x5()));
                        main.setImage(img);
                    }
                });
        rm.add(
                new RunMenuItem("Median Cross 7x7") {
                    public void run() {
                        Image img = main.getImage();
                        img
                                = ImageUtils.convolution(
                                        img,
                                        shortMatrixToFloat(
                                                Kernels.getMedianCross7x7()));
                        main.setImage(img);
                    }
                });
        rm.add(
                new RunMenuItem("Median Diamond") {
                    public void run() {
                        Image img = main.getImage();
                        img
                                = ImageUtils.convolution(
                                        img,
                                        shortMatrixToFloat(
                                                Kernels.getMedianDiamond()));
                        main.setImage(img);
                    }
                });
        rm.add(
                new RunMenuItem("Median Octagon 5x5") {
                    public void run() {
                        Image img = main.getImage();
                        img
                                = ImageUtils.convolution(
                                        img,
                                        shortMatrixToFloat(
                                                Kernels.getMedianOctagon5x5()));
                        main.setImage(img);
                    }
                });
        rm.add(
                new RunMenuItem("Median Square 7x7") {
                    public void run() {
                        Image img = main.getImage();
                        img
                                = ImageUtils.convolution(
                                        img,
                                        shortMatrixToFloat(
                                                Kernels.getMedianSquare7x7()));
                        main.setImage(img);
                    }
                });


        return rm;
    }
}
